from __future__ import with_statement
import os.path
import logging
import logging.handlers
from nas.system import get_server_domain

MAIN_CONF_FILE = '/etc/sockproxy/sockproxy.conf'
CACHED_SERVER_FILE = '/var/lib/sockproxy/server'

DEFAULT_LOCAL_TARGET_PORT = 80
DEFAULT_CA_FILE = '/etc/sockproxy/ca.crt'

DEFAULT_CONNECT_TRY_TIMES = 3
DEFAULT_INITIAL_RECONNECT_DELAY = 5
DEFAULT_MAX_RECONNECT_DELAY = 60


def get_host_port(addr, default_port=80):
    """docstring for get_host_p"""
    tmp = addr.partition(':')
    return (tmp[0], (tmp[2] and int(tmp[2]) or default_port) )

def _get_absolute_path(conffile):
    if conffile[0] == '/':
        return conffile
    else:
        return os.path.join('/etc/sockproxy', conffile)

class _Config(object):
    """Sock proxy config object.
    It is not supposed to be used directly. Use the global config instance instead
    """
    def __init__(self):
        super(_Config, self).__init__()
        self.default_server = None
        self.cached_server = None
        self.ca_file = DEFAULT_CA_FILE
        self.connect_try_times = DEFAULT_CONNECT_TRY_TIMES
        self.initial_reconnect_delay = DEFAULT_INITIAL_RECONNECT_DELAY
        self.max_reconnect_delay = DEFAULT_MAX_RECONNECT_DELAY
        self.properties = {}
        self._load_config()
        try:
            self._load_cached_server()
        except:
            pass

    def _load_config(self):
        """Load configuration from config file.
        The format is:
            propertyname=value
        """
        with open(MAIN_CONF_FILE, 'r') as f:
            for line in f:
                p = line.find('#')
                if p > 0:
                    line = line[0:p]
                name, _, value = line.partition('=')
                name = name.strip()
                value = value.strip()
                self.properties[name] = value
        self.default_server = (
                get_server_domain(),
                int(self.properties['default_port']),
                )
        if 'ca_file' in self.properties:
            self.ca_file = _get_absolute_path(self.properties['ca_file'])

    def _load_cached_server(self):
        """Load cached server info.
        The data format is:
            host:port:server_id
        """
        with open(CACHED_SERVER_FILE, 'rw') as f:
            host, port, serverid = file(CACHED_SERVER_FILE).read().strip().split(":")[:3]
            self.cached_server = (host, int(port), serverid)
#            for line in f:
#                line = line.strip()
#                if line:
#                    host, port
#                    host, _, other = line.partition(':')
#                    port, _, serverid = other.partition(':')
#                    self.cached_server = (host, int(port), serverid)
#                    break
    def update_status_cache(self, old, status):
        "Update status info"
        if status != old:
            ip, port,serverid = file(CACHED_SERVER_FILE).read().strip().split(':')[:3]
            with open(CACHED_SERVER_FILE, 'w') as f:
                f.write(":".join((ip, port, serverid, str(status))))
            old = status

    def update_cached_server(self, host, port, serverid):
        """Update cached server info"""
        assert host, "Host is empty"
        assert port and isinstance(port, int), "Invalid port"
        assert serverid, "Serverid is empty"
        self.cached_server = (host, port, serverid)
        with open(CACHED_SERVER_FILE, 'w') as f:
            f.write("%s:%s:%s\n" % (host, port, serverid))

'''
The config instance, should use this one
'''
config = _Config()

levels = {
        'debug' : logging.DEBUG,
        'info' : logging.INFO,
        'warning' : logging.WARNING,
        'error' : logging.ERROR,
        'critical' : logging.CRITICAL
    }

#Config logging
def config_logging():
    """logging configuration"""
    logfile = '/var/log/sockproxy.log'
    level = logging.INFO
    if 'log' in config.properties:
        logfile = config.properties['log']
    if 'log_level' in config.properties:
        level = levels.get(config.properties['log_level'].lower(), logging.INFO)

    # Only configure the root logger
    logger = logging.getLogger()
    logger.setLevel(level)

    handler = logging.handlers.RotatingFileHandler(logfile, maxBytes=204800, backupCount=5)
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def config_logging_debug():
    """this logging configuration is intended only for testing"""
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(levelname)s %(message)s")
