from __future__ import with_statement

import socket
import ssl
import os
import os.path
import errno
import time
import logging
from sockproxy.tornado import iostream, ioloop
from sockproxy.config import config
from sockproxy.config import get_host_port
from sockproxy import protocol
from nas.info import getseckey
from nas.network import gethwaddr

DEFAULT_RECONNECT_DELAY = 5

DEFAULT_CONTROL_PORT = 9000
DEFAULT_KEEPALIVE = 27
TIMEOUT_THRESHOLD = 30
DEFAULT_TIMEOUT = DEFAULT_KEEPALIVE + TIMEOUT_THRESHOLD

PROXY_MAX_READ_SIZE = 8192

ALARM_FILE = '/var/lib/sockproxy/alarm'

_get_keep_alive = lambda params : int(params['keep_alive'].strip().split()[0])

class ControlConnection(object):

    def __init__(self, host, port, ssl_options, connection_manager, last_server_id=None, keep_alive=None):
        super(ControlConnection, self).__init__()
        self.status = False
        self.host = host
        self.port = port
        self._connect_manager = connection_manager
        self._last_server_id = last_server_id
        self._keep_alive = keep_alive
        config.update_status_cache(self.status, False)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._iostream = iostream.SSLIOStream(sock, ssl_options=ssl_options)
        self._iostream.set_close_callback(self._on_close)
        if keep_alive:
            self.reconnect_deadline = time.time() + keep_alive + TIMEOUT_THRESHOLD
        else:
            self.reconnect_deadline = time.time() + DEFAULT_TIMEOUT

    def start(self):
        logging.info("Connecting to %s:%s" % (self.host, self.port))
        try:
            self._iostream.connect((self.host, self.port), self._nas_online_request)
        except socket.error, e:
            config.update_status_cache(self.status, False)
            logging.exception("Error connecting to %s:%s" % (self.host, self.port))
            self.close()

    def close(self):
        config.update_status_cache(self.status, False)
        self._iostream.close()

    def _on_close(self):
        """Close callback for control connection, for logging """
        config.update_status_cache(self.status, False)
        logging.info("Control connection closed/failed, will try to connect again in a few seconds")
        self._connect_manager.retry_connect_to_server()

    def _nas_online_request(self):
        err = self._iostream.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err != 0:
            logging.error("Error connecting to %s:%s, errno:%s" % (self.host, self.port, err))
            self.close()
            return
        regkey = gethwaddr().replace(':', '')
        sec = getseckey()
        logging.info('Sending nas-online message')
        protocol.nas_online(self._iostream,
                            regkey,
                            sec,
                            redirected_list=self._connect_manager.redirected_list,
                            last_server_id=self._last_server_id,
                            keep_alive=self._keep_alive)
        protocol.read_response(self._iostream, self._nas_online_response)

    def _nas_online_response(self, status, params):
        logging.info('nas-online response (status %s, params %s)' % (status, str(params)) )
        if status == 200:
            server_id = params.get('server_id')
            host, port = get_host_port(params.get('server_addr'))
            self._connect_manager.update_cached_server(host, port, server_id)
            config.update_status_cache(self.status, True)
            self._update_keep_alive(_get_keep_alive(params))
            protocol.read_server_command(self._iostream, self._handle_server_command) #Start waiting for server's command
        elif status == 301:
            host, port = get_host_port(params['location'], default_port=DEFAULT_CONTROL_PORT)
            '''
            Setting close callback to None before redirecting, to avoid this connection
            being retried.
            '''
            self._iostream.set_close_callback(None)
            self._connect_manager.redirect(host, port)
        else:
            logging.error('Server returned a bad status code %s, closing connection' % status)
            self.close()

    def _handle_server_command(self, params):
        logging.debug('server command received (params %s)' % str(params))
        if 'connect_back' in params:
            self._connect_back(params)
        elif 'keep_alive' in params:
            logging.debug('Keep alive received')
            self._update_keep_alive(_get_keep_alive(params))
        else:
            logging.warn('Invalid command received: ' + str(params))

        protocol.read_server_command(self._iostream, self._handle_server_command)

    def _connect_back(self, params):
        host, port = get_host_port(params.get('connect_back', ''))
        local_host, local_port = get_host_port(params.get('service', ''))
        cookie = params.get('cookie', '')
        logging.debug('Connecting back to %s:%s, and bridge local service at %s:%s' % (host, port, local_host, local_port))

        use_ssl = False
        if params.get('ssl', '') == 'true':
            use_ssl  = True
        self._connect_manager.connect_back(host, port, local_host, local_port, cookie, use_ssl=use_ssl)
        self._update_keep_alive(_get_keep_alive(params))

    def _update_keep_alive(self, keep_alive):
        self._connect_manager.update_keep_alive(keep_alive)
        timeout = keep_alive + TIMEOUT_THRESHOLD
        self._update_reconnect_deadline(timeout)

    def _update_reconnect_deadline(self, timeout):
        self.reconnect_deadline = time.time() + timeout


class ProxyConnection(object):
    """Proxy connection that bridges a local lighttpd connection with a remote connection"""

    def __init__(self, host, port, local_host, local_port, connect_back_cookie, connection_manager, ssl_options=None):
        super(ProxyConnection, self).__init__()
        self.host = host
        self.port = port
        self.local_host = local_host
        self.local_port = local_port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if ssl_options:
            self._iostream = iostream.SSLIOStream(sock, ssl_options=ssl_options)
        else:
            self._iostream = iostream.IOStream(sock)
        self._iostream.set_close_callback(self._on_serverconn_close)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._local_iostream = iostream.IOStream(sock)
        self._local_iostream.set_close_callback(self._on_proxyconn_close)

        self._connect_back_cookie = connect_back_cookie
        self._connect_manager = connection_manager

        self._server_ready = False
        self._local_ready = False

    def start(self):
        logging.debug("Starting proxy connection, server host %s, port %s. Local port %s" % (self.host, self.port, self.local_port))
        try:
            self._iostream.connect((self.host, self.port), self._on_server_connected)
            self._local_iostream.connect((self.local_host, self.local_port), self._on_local_connected)
        except socket.error, e:
            logging.exception("Error creating proxy connection")
            self.close()

    def close(self):
        if not self._iostream.closed():
            self._iostream.close()
        if not self._local_iostream.closed():
            self._local_iostream.close()

    def _on_server_connected(self):
        logging.debug("Sending connect-back cookie: %s" % self._connect_back_cookie)
        protocol.send_connect_back_cookie(self._iostream, self._connect_back_cookie)
        self._server_ready = True
        self._start_bridging()

    def _on_local_connected(self):
        self._local_ready = True
        self._start_bridging()

    def _start_bridging(self):
        # Start brigding only when both ends are ready
        if self._local_ready and self._server_ready:
            self._read_server_end()
            self._read_local_end()

    def _read_server_end(self):
        self._iostream.read_max_bytes(PROXY_MAX_READ_SIZE, self._send_to_local_end)

    def _send_to_local_end(self, data):
        self._local_iostream.write(data, self._read_server_end)

    def _read_local_end(self):
        self._local_iostream.read_max_bytes(PROXY_MAX_READ_SIZE, self._send_to_server_end)

    def _send_to_server_end(self, data):
        self._iostream.write(data, self._read_local_end)

    def _on_proxyconn_close(self):
        logging.debug('Proxy connection closed')
        if not self._iostream.closed():
            self._try_close_server_connection()

    def _try_close_server_connection(self):
        """Try to close server connection. If it is in writing status, then try later"""
        if self._iostream.closed():
            return
        if self._iostream.writing():
            logging.debug('Server connection stilling writing, try closing later')
            self._connect_manager.schedule_task(self._try_close_server_connection)
        else:
            self._iostream.close()
            self._connect_manager.remove_proxy_conn(self)

    def _on_serverconn_close(self):
        logging.debug('Server connection closed')
        if not self._local_iostream.closed():
            self._local_iostream.close()
        self._connect_manager.remove_proxy_conn(self)


class ConnectionManager(object):

    RECONNECT_DELAY_STEP = 5

    def __init__(self):
        super(ConnectionManager, self).__init__()
        self.redirected_list = set()
        self._cached_server_connect_times = 0
        self._redirect_target = None
        self._redirect_connect_times = 0
        self._redirecting = False
        self._control_conn = None
        self._proxy_conns = set()
        self._ioloop = ioloop.IOLoop.instance()
        self._keep_alive = None
        self._valid_keep_alive = None
        self._longest_keep_alive = None
        self._reconnect_delay = config.initial_reconnect_delay

    def start(self):
        self.connect_to_server()
        self._ioloop.start()

    def stop(self):
        logging.info('Stopping connection manager')
        if self._control_conn:
            self._control_conn.close()
        for conn in self._proxy_conns:
            conn.close()
        self._ioloop.stop()

    def connect_to_server(self):
        server_id = None
        if self._redirecting and self._redirect_connect_times < config.connect_try_times:
            host, port = self._redirect_target
            self._redirect_connect_times = self._redirect_connect_times + 1
            logging.debug('Try redirected server %s:%s for the %s time(s)' % (host, port, self._redirect_connect_times))
        elif config.cached_server and self._cached_server_connect_times < config.connect_try_times:
            host, port, server_id = config.cached_server
            self._cached_server_connect_times =  self._cached_server_connect_times + 1
            logging.debug('Try connecting cached server %s:%s for the %s time(s). Server ID: %s' %
                    (host, port, self._cached_server_connect_times, server_id))
        else:
            host, port = config.default_server
            self._redirecting = False
            self._redirect_connect_times = 0
            self.redirected_list.clear()
            logging.debug('Connecting to default server %s:%s' % (host, port))

        self._control_conn = ControlConnection(host,
                                               port,
                                               {'ca_certs': config.ca_file,
                                                'cert_reqs': ssl.CERT_REQUIRED},
                                               self,
                                               last_server_id=server_id,
                                               keep_alive=self._longest_keep_alive)
        self._control_conn.start()
        self._ioloop.add_timeout(self._control_conn.reconnect_deadline, self._check_keep_alive)

    def retry_connect_to_server(self, delaysecs=DEFAULT_RECONNECT_DELAY):
        self._control_conn = None
        deadline = time.time() + self._reconnect_delay
        self._ioloop.add_timeout(deadline, self.connect_to_server)
        # Update reconnect delay
        self._reconnect_delay += self.RECONNECT_DELAY_STEP
        if self._reconnect_delay > config.max_reconnect_delay:
            self._reconnect_delay = config.max_reconnect_delay

    def redirect(self, host, port):
        logging.debug('Redirecting to %s:%s' % (host, port))
        self._redirecting = True
        self._redirect_target = (host, port)
        self._redirect_connect_times = 0
        self.redirected_list.add((self._control_conn.host, self._control_conn.port))
        if self._control_conn:
            self._control_conn.close()
            self._control_conn = None
        self.connect_to_server()

    def connect_back(self, host, port, local_host, local_port, cookie, use_ssl=False):
        if use_ssl:
            ssl_options = {'ca_certs': config.ca_file,
                           'cert_reqs': ssl.CERT_REQUIRED}
        else:
            ssl_options = None
        proxy_conn = ProxyConnection(host,
                                     port,
                                     local_host,
                                     local_port,
                                     cookie,
                                     self,
                                     ssl_options=ssl_options)
        self._proxy_conns.add(proxy_conn)
        proxy_conn.start()

    def remove_proxy_conn(self, proxy_conn):
        if proxy_conn in self._proxy_conns:
            self._proxy_conns.remove(proxy_conn)

    def schedule_task(self, callback):
        self._ioloop.add_callback(callback)

    def update_cached_server(self, host, port, server_id):
        config.update_cached_server(host, port, server_id)
        self._redirecting = False
        self._redirect_connect_times = 0
        self._cached_server_connect_times = 0 # Connect succeeded, so we reset the connection counter
        self.redirected_list.clear()

    def update_keep_alive(self, keep_alive):
        '''
        We get a keep alive value that is longer than previously saved one. In this case,
        we will save this value in _valid_keep_alive, and if a timeout happens, we use this
        value as the keep-alive interval in a new connection.
        '''
        if self._keep_alive and keep_alive > self._keep_alive:
            self._valid_keep_alive = self._keep_alive
            self._clear_alarm()
            # Reset reconnect delay
            self._reconnect_delay = config.initial_reconnect_delay
        self._keep_alive = keep_alive

    def _check_keep_alive(self):
        if not self._control_conn: #If current control connection is closed
            return
        current_time = time.time()
        if current_time >= self._control_conn.reconnect_deadline:
            logging.info('Control connection keep-alive timed out, close and reconnect...')
            self._control_conn.close()
            '''
            _valid_keep_alive is None means the client does not received any
            keep-alive command. In this case, the keep-alive time is the shortest
            value that the server could provide, and we need to generate an alarm to
            inform the user to check the network environment.
            '''
            if self._valid_keep_alive:
                self._longest_keep_alive = self._valid_keep_alive
            else:
                self._generate_alarm()
        else:
            self._ioloop.add_timeout(self._control_conn.reconnect_deadline, self._check_keep_alive)

    def _generate_alarm(self):
        """An alarm is generated if the connection timed out even we use the smallest keep-alive time"""
        with open(ALARM_FILE, 'w') as f:
            logging.debug('Generate alarm')
            f.write('''Need to check network for socket proxy to work properly.  Current keep alive time: %s\n'''
                    % self._keep_alive)

    def _clear_alarm(self):
        """Clear the alarm previously generated if it exists"""
        if os.path.exists(ALARM_FILE):
            logging.debug('Clear alarm')
            os.unlink(ALARM_FILE)

